@@ -1,19 +1,22 @@
 import polars as pl
 import re
-from typing import Literal, Union, Optional, Any, Callable, List, Dict
-from .utilities import _script_info
-import pandas as pd
+from typing import Literal, Union, Optional, Any, Callable, List, Dict, Tuple
+from ._script_info import _script_info
+from ._logger import _LOGGER
+import warnings
 
 
 __all__ = [
     "ColumnCleaner",
-    "DataFrameCleaner"
+    "DataFrameCleaner",
     "TransformationRecipe",
     "DataProcessor",
+    "BinaryTransformer",
+    "MultiBinaryDummifier",
     "KeywordDummifier",
     "NumberExtractor",
     "MultiNumberExtractor",
-    "RatioCalculator"
+    "RatioCalculator",
     "CategoryMapper",
     "RegexMapper",
     "ValueBinner",
@@ -21,100 +24,141 @@ __all__ = [
 ]
 
 ########## EXTRACT and CLEAN ##########
-
 class ColumnCleaner:
     """
-    Cleans and standardizes a single pandas Series based on a dictionary of regex-to-value replacement rules.
+    A configuration object that defines cleaning rules for a single Polars DataFrame column.
+
+    This class holds a dictionary of regex-to-replacement rules, the target column name,
+    and the case-sensitivity setting. It is intended to be used with the DataFrameCleaner.
     
+    Notes:
+        - Define rules from most specific to more general to create a fallback system.
+        - Beware of chain replacements (rules matching strings that have already been
+          changed by a previous rule in the same cleaner).
+
     Args:
+        column_name (str):
+            The name of the column to be cleaned.
         rules (Dict[str, str]):
-            A dictionary where each key is a regular expression pattern and
-            each value is the standardized string to replace matches with.
+            A dictionary of regex patterns to replacement strings. Can use
+            backreferences (e.g., r'$1 $2') for captured groups. Note that Polars
+            uses a '$' prefix for backreferences.
+        case_insensitive (bool):
+            If True (default), regex matching ignores case.
+
+    ## Usage Example
+
+    ```python
+    id_rules = {
+        # Matches 'ID-12345' or 'ID 12345' and reformats to 'ID:12345'
+        r'ID[- ](\\d+)': r'ID:$1'
+    }
+
+    id_cleaner = ColumnCleaner(column_name='user_id', rules=id_rules)
+    # This object would then be passed to a DataFrameCleaner.
+    ```
     """
-    def __init__(self, rules: Dict[str, str]):
+    def __init__(self, column_name: str, rules: Dict[str, str], case_insensitive: bool = True):
+        if not isinstance(column_name, str) or not column_name:
+            raise TypeError("The 'column_name' must be a non-empty string.")
         if not isinstance(rules, dict):
             raise TypeError("The 'rules' argument must be a dictionary.")
 
-        # Validate that all keys are valid regular expressions
+        # Validate each regex pattern for correctness
         for pattern in rules.keys():
             try:
                 re.compile(pattern)
             except re.error as e:
                 raise ValueError(f"Invalid regex pattern '{pattern}': {e}") from e
 
+        self.column_name = column_name
         self.rules = rules
-
-    def clean(self, series: pd.Series) -> pd.Series:
-        """
-        Applies the standardization rules to the provided Series (requires string data).
-        
-        Non-matching values are kept as they are.
-
-        Args:
-            series (pd.Series): The pandas Series to clean.
-
-        Returns:
-            pd.Series: A new Series with the values cleaned and standardized.
-        """
-        return series.astype(str).replace(self.rules, regex=True)
+        self.case_insensitive = case_insensitive
 
 
 class DataFrameCleaner:
     """
-    Orchestrates the cleaning of multiple columns in a pandas DataFrame using a nested dictionary of rules and `ColumnCleaner` objects.
+    Orchestrates cleaning multiple columns in a Polars DataFrame.
+
+    This class takes a list of ColumnCleaner objects and applies their defined
+    rules to the corresponding columns of a DataFrame using high-performance
+    Polars expressions.
 
     Args:
-        rules (Dict[str, Dict[str, str]]):
-            A nested dictionary where each top-level key is a column name,
-            and its value is a dictionary of regex rules for that column, as expected by `ColumnCleaner`.
-    """
-    def __init__(self, rules: Dict[str, Dict[str, str]]):
-        if not isinstance(rules, dict):
-            raise TypeError("The 'rules' argument must be a nested dictionary.")
-        
-        for col_name, col_rules in rules.items():
-            if not isinstance(col_rules, dict):
-                raise TypeError(
-                    f"The value for column '{col_name}' must be a dictionary "
-                    f"of rules, but got type {type(col_rules).__name__}."
-                )
-        
-        self.rules = rules
+        cleaners (List[ColumnCleaner]):
+            A list of ColumnCleaner configuration objects.
 
-    def clean(self, df: pd.DataFrame) -> pd.DataFrame:
+    Raises:
+        TypeError: If 'cleaners' is not a list or contains non-ColumnCleaner objects.
+        ValueError: If multiple ColumnCleaner objects target the same column.
+    """
+    def __init__(self, cleaners: List[ColumnCleaner]):
+        if not isinstance(cleaners, list):
+            raise TypeError("The 'cleaners' argument must be a list of ColumnCleaner objects.")
+
+        seen_columns = set()
+        for cleaner in cleaners:
+            if not isinstance(cleaner, ColumnCleaner):
+                raise TypeError(
+                    f"All items in 'cleaners' list must be ColumnCleaner objects, "
+                    f"but found an object of type {type(cleaner).__name__}."
+                )
+            if cleaner.column_name in seen_columns:
+                raise ValueError(
+                    f"Duplicate ColumnCleaner found for column '{cleaner.column_name}'. "
+                    "Each column should only have one cleaner."
+                )
+            seen_columns.add(cleaner.column_name)
+
+        self.cleaners = cleaners
+
+    def clean(self, df: pl.DataFrame) -> pl.DataFrame:
         """
-        Applies all defined cleaning rules to the DataFrame.
+        Applies all defined cleaning rules to the Polars DataFrame.
 
         Args:
-            df (pd.DataFrame): The pandas DataFrame to clean.
+            df (pl.DataFrame): The Polars DataFrame to clean.
 
         Returns:
-            pd.DataFrame: A new, cleaned DataFrame.
+            pl.DataFrame: A new, cleaned Polars DataFrame.
+
+        Raises:
+            ValueError: If any columns specified in the cleaners are not found
+                        in the input DataFrame.
         """
-        rule_columns = set(self.rules.keys())
+        rule_columns = {c.column_name for c in self.cleaners}
         df_columns = set(df.columns)
-        
         missing_columns = rule_columns - df_columns
-        
+
         if missing_columns:
-            # Report all missing columns in a single, clear error message
             raise ValueError(
-                f"The following columns specified in the cleaning rules "
+                f"The following columns specified in cleaning rules "
                 f"were not found in the DataFrame: {sorted(list(missing_columns))}"
             )
+
+        df_cleaned = df.clone()
         
-        # Start the process
-        df_cleaned = df.copy()
-        
-        for column_name, column_rules in self.rules.items():
-            # Create and apply the specific cleaner for the column
-            cleaner = ColumnCleaner(rules=column_rules)
-            df_cleaned[column_name] = cleaner.clean(df_cleaned[column_name])
+        # Build and apply a series of expressions for each column
+        for cleaner in self.cleaners:
+            col_name = cleaner.column_name
+            
+            # Start with the column, cast to String for replacement operations
+            col_expr = pl.col(col_name).cast(pl.String)
+
+            # Sequentially chain 'replace_all' expressions for each rule
+            for pattern, replacement in cleaner.rules.items():
+                final_pattern = f"(?i){pattern}" if cleaner.case_insensitive else pattern
+                col_expr = col_expr.str.replace_all(final_pattern, replacement)
+            
+            # Execute the expression chain for the column
+            df_cleaned = df_cleaned.with_columns(col_expr.alias(col_name))
+            
+        print(f"Cleaned {len(self.cleaners)} columns.")
             
         return df_cleaned
 
 
-############ TRANSFORM ####################
+############ TRANSFORM MAIN ####################
 
 # Magic word for rename-only transformation
 _RENAME = "rename"
@@ -251,7 +295,7 @@ class DataProcessor:
                 raise TypeError(f"Invalid 'transform' action for '{input_col_name}': {transform_action}")
 
         if not processed_columns:
-            print("Warning: The transformation resulted in an empty DataFrame.")
+            _LOGGER.warning("⚠️ The transformation resulted in an empty DataFrame.")
             return pl.DataFrame()
             
         return pl.DataFrame(processed_columns)
@@ -299,13 +343,142 @@ class DataProcessor:
         """
         print(self)
 
+############ TRANSFORMERS ####################
+
+class BinaryTransformer:
+    """
+    A transformer that maps string values to a binary 1 or 0 based on keyword matching.
+
+    Must supply a list of keywords for either the 'true' case (1) or the 'false' case (0), but not both.
+
+    Args:
+        true_keywords (List[str] | None):
+            If a string contains any of these keywords, the output is 1, otherwise 0.
+        false_keywords (List[str] | None):
+            If a string contains any of these keywords, the output is 0, otherwise 1.
+    """
+    def __init__(
+        self,
+        true_keywords: Optional[List[str]] = None,
+        false_keywords: Optional[List[str]] = None,
+        case_insensitive: bool = True,
+    ):
+        # --- Validation: Enforce one and only one option ---
+        if true_keywords is not None and false_keywords is not None:
+            raise ValueError(
+                "Provide either 'true_keywords' or 'false_keywords', but not both."
+            )
+        if true_keywords is None and false_keywords is None:
+            raise ValueError(
+                "You must provide either 'true_keywords' or 'false_keywords'."
+            )
+
+        # --- Configuration ---
+        self.keywords: List[str] = true_keywords if true_keywords is not None else false_keywords # type: ignore
+        if not self.keywords:
+            raise ValueError("Keyword list cannot be empty.")
+
+        self.mode: str = "true_mode" if true_keywords is not None else "false_mode"
+        
+        # --- Create the regex string pattern ---
+        # Escape keywords to treat them as literals
+        base_pattern = "|".join(re.escape(k) for k in self.keywords)
+
+        # For polars, add case-insensitivity flag `(?i)` to the pattern string itself
+        if case_insensitive:
+            self.pattern = f"(?i){base_pattern}"
+        else:
+            self.pattern = base_pattern
+        
+
+    def __call__(self, column: pl.Series) -> pl.Series:
+        """
+        Applies the binary mapping logic to the input column.
+
+        Args:
+            column (pl.Series): The input Polars Series of string data.
+
+        Returns:
+            pl.Series: A new Series of type UInt8 containing 1s and 0s.
+        """
+        # Create a boolean Series: True if any keyword is found, else False
+        contains_keyword = column.str.contains(self.pattern)
+
+        # Apply logic and cast directly to integer type
+        if self.mode == "true_mode":
+            # True -> 1, False -> 0
+            return contains_keyword.cast(pl.UInt8)
+        else: # false_mode
+            # We want the inverse: True -> 0, False -> 1
+            return (~contains_keyword).cast(pl.UInt8)
+
+
+class MultiBinaryDummifier:
+    """
+    A one-to-many transformer that creates multiple binary columns from a single
+    text column based on a list of keywords.
+
+    For each keyword provided, this transformer generates a corresponding column
+    with a value of 1 if the keyword is present in the input string, and 0 otherwise.
+    It is designed to be used within the DataProcessor pipeline.
+
+    Args:
+        keywords (List[str]):
+            A list of strings, where each string is a keyword to search for. A separate
+            binary column will be created for each keyword.
+        case_insensitive (bool):
+            If True, keyword matching ignores case. Defaults to True.
+    """
+    def __init__(self, keywords: List[str], case_insensitive: bool = True):
+        if not isinstance(keywords, list) or not all(isinstance(k, str) for k in keywords):
+            raise TypeError("The 'keywords' argument must be a list of strings.")
+        if not keywords:
+            raise ValueError("The 'keywords' list cannot be empty.")
+
+        self.keywords = keywords
+        self.case_insensitive = case_insensitive
+
+    def __call__(self, column: pl.Series) -> pl.DataFrame:
+        """
+        Executes the dummification logic.
+
+        Args:
+            column (pl.Series): The input Polars Series to transform.
+
+        Returns:
+            pl.DataFrame: A DataFrame where each column corresponds to a keyword.
+        """
+        # Ensure the input is treated as a string, preserving nulls
+        str_column = column.cast(pl.Utf8)
+        
+        output_expressions = []
+        for i, keyword in enumerate(self.keywords):
+            # Escape keyword to treat it as a literal, not a regex pattern
+            base_pattern = re.escape(keyword)
+
+            # Add case-insensitivity flag `(?i)` if needed
+            pattern = f"(?i){base_pattern}" if self.case_insensitive else base_pattern
+
+            # Create the binary expression
+            expr = (
+                pl.when(str_column.is_null())
+                .then(None) # Propagate nulls from original column
+                .when(str_column.str.contains(pattern))
+                .then(pl.lit(1, dtype=pl.UInt8))
+                .otherwise(pl.lit(0, dtype=pl.UInt8))
+                .alias(f"col_{i}") # Generic name for DataProcessor
+            )
+            output_expressions.append(expr)
+
+        return pl.select(output_expressions)
+
 
 class KeywordDummifier:
     """
     A configurable transformer that creates one-hot encoded columns based on
     keyword matching in a Polars Series.
 
-    Instantiate this class with keyword configurations. The instance can be used as a 'transform' callable compatible with the `TransformationRecipe`.
+    Operates on a "first match wins" principle.
 
     Args:
         group_names (List[str]): 
@@ -315,13 +488,16 @@ class KeywordDummifier:
         group_keywords (List[List[str]]): 
             A list of lists of strings. Each inner list corresponds to a 
             `group_name` at the same index and contains the keywords to search for.
+        case_insensitive (bool):
+            If True, keyword matching ignores case.
     """
-    def __init__(self, group_names: List[str], group_keywords: List[List[str]]):
+    def __init__(self, group_names: List[str], group_keywords: List[List[str]], case_insensitive: bool = True):
         if len(group_names) != len(group_keywords):
             raise ValueError("Initialization failed: 'group_names' and 'group_keywords' must have the same length.")
         
         self.group_names = group_names
         self.group_keywords = group_keywords
+        self.case_insensitive = case_insensitive
 
     def __call__(self, column: pl.Series) -> pl.DataFrame:
         """
@@ -335,16 +511,25 @@ class KeywordDummifier:
         """
         column = column.cast(pl.Utf8)
         
-        categorize_expr = pl.when(pl.lit(False)).then(pl.lit(None))
+        categorize_expr = pl.when(pl.lit(False)).then(pl.lit(None, dtype=pl.Utf8))
+        
         for name, keywords in zip(self.group_names, self.group_keywords):
-            pattern = "|".join(re.escape(k) for k in keywords)
+            # Create the base regex pattern by escaping and joining keywords
+            base_pattern = "|".join(re.escape(k) for k in keywords)
+            
+            # Add the case-insensitive flag `(?i)` to the pattern string
+            if self.case_insensitive:
+                pattern = f"(?i){base_pattern}"
+            else:
+                pattern = base_pattern
+            
             categorize_expr = categorize_expr.when(
                 column.str.contains(pattern)
             ).then(pl.lit(name))
         
         categorize_expr = categorize_expr.otherwise(None).alias("category")
 
-        temp_df = pl.DataFrame(categorize_expr)
+        temp_df = pl.select(categorize_expr)
         df_with_dummies = temp_df.to_dummies(columns=["category"])
         
         final_columns = []
@@ -356,9 +541,10 @@ class KeywordDummifier:
                     df_with_dummies.get_column(dummy_col_name).alias(name)
                 )
             else:
+                # If a group had no matches, create a column of zeros
                 final_columns.append(pl.lit(0, dtype=pl.UInt8).alias(name))
 
-        return pl.DataFrame(final_columns)
+        return pl.select(final_columns)
 
 
 class NumberExtractor:
@@ -383,7 +569,7 @@ class NumberExtractor:
         self,
         regex_pattern: str = r"(\d+\.?\d*)",
         dtype: Literal["float", "int"] = "float",
-        round_digits: Optional[int] = None,
+        round_digits: Optional[int] = 2,
     ):
         # --- Validation ---
         if not isinstance(regex_pattern, str):
@@ -403,7 +589,7 @@ class NumberExtractor:
             if not isinstance(round_digits, int):
                 raise TypeError("round_digits must be an integer.")
             if dtype == "int":
-                print(f"Warning: 'round_digits' is specified but dtype is 'int'. Rounding will be ignored.")
+                _LOGGER.warning(f"⚠️ 'round_digits' is specified but dtype is 'int'. Rounding will be ignored.")
 
         self.regex_pattern = regex_pattern
         self.dtype = dtype
@@ -515,26 +701,28 @@ class MultiNumberExtractor:
 
 class RatioCalculator:
     """
-    A transformer that parses a string ratio (e.g., "40:5" or "30/2") and computes the result of the division.
-
-    Args:
-        regex_pattern (str, optional):
-            The regex pattern to find the numerator and denominator. It MUST
-            contain exactly two capturing groups: the first for the
-            numerator and the second for the denominator. Defaults to a
-            pattern that handles common delimiters like ':' and '/'.
+    A transformer that parses a string ratio (e.g., "40:5" or "30/2") and
+    computes the result of the division. It gracefully handles strings that
+    do not match the pattern by returning null.
     """
     def __init__(
         self,
-        regex_pattern: str = r"(\d+\.?\d*)\s*[:/]\s*(\d+\.?\d*)"
+        # Default pattern includes the full-width colon '：'
+        regex_pattern: str = r"(\d+\.?\d*)\s*[:：/]\s*(\d+\.?\d*)"
     ):
-        # --- Validation ---
+        # --- Robust Validation ---
         try:
-            if re.compile(regex_pattern).groups != 2:
+            compiled_pattern = re.compile(regex_pattern)
+            if compiled_pattern.groups != 2:
                 raise ValueError(
-                    "regex_pattern must contain exactly two "
+                    "RatioCalculator regex_pattern must contain exactly two "
                     "capturing groups '(...)'."
                 )
+            if compiled_pattern.groupindex:
+                raise ValueError(
+                    "RatioCalculator must be initialized with unnamed capturing groups "
+                    "(e.g., '(\\d+)'), not named groups (e.g., '(?P<name>\\d+)')."
+                )
         except re.error as e:
             raise ValueError(f"Invalid regex pattern provided: {e}") from e
 
@@ -543,27 +731,20 @@ class RatioCalculator:
     def __call__(self, column: pl.Series) -> pl.Series:
         """
         Applies the ratio calculation logic to the input column.
-
-        Args:
-            column (pl.Series): The input Polars Series of ratio strings.
-
-        Returns:
-            pl.Series: A new Series of floats containing the division result.
-                       Returns null for invalid formats or division by zero.
+        This version uses .str.extract() for maximum stability.
         """
-        # .extract_groups returns a struct with a field for each capture group
-        # e.g., {"group_1": "40", "group_2": "5"}
-        groups = column.str.extract_groups(self.regex_pattern)
+        # Extract numerator (group 1) and denominator (group 2) separately.
+        numerator_expr = column.str.extract(self.regex_pattern, 1).cast(pl.Float64, strict=False)
+        denominator_expr = column.str.extract(self.regex_pattern, 2).cast(pl.Float64, strict=False)
 
-        # Extract numerator and denominator, casting to float
-        # strict=False ensures that non-matches become null
-        numerator = groups.struct.field("group_1").cast(pl.Float64, strict=False)
-        denominator = groups.struct.field("group_2").cast(pl.Float64, strict=False)
+        # Calculate the ratio, handling division by zero.
+        final_expr = pl.when(denominator_expr != 0).then(
+            numerator_expr / denominator_expr
+        ).otherwise(
+            None # Handles both null denominators and division by zero
+        )
 
-        # Safely perform division, returning null if denominator is 0
-        return pl.when(denominator != 0).then(
-            numerator / denominator
-        ).otherwise(None)
+        return pl.select(final_expr.round(4)).to_series()
 
 
 class CategoryMapper:
@@ -631,33 +812,42 @@ class RegexMapper:
     "first match wins" logic makes the order of the mapping important.
 
     Args:
-        mapping (Dict[str, Union[int, float]]):
+        mapping (Dict[str, [int | float]]):
             An ordered dictionary where keys are regex patterns and values are
             the numbers to map to if the pattern is found.
-        unseen_value (Optional[Union[int, float]], optional):
+        unseen_value (int | float | None):
             The numerical value to use for strings that do not match any
-            of the regex patterns. If None (default), unseen values are
-            mapped to null.
+            of the regex patterns. If None, unseen values are mapped to null.
+        case_insensitive (bool):
+            If True , the regex matching for all patterns will ignore case.
     """
     def __init__(
         self,
         mapping: Dict[str, Union[int, float]],
         unseen_value: Optional[Union[int, float]] = None,
+        case_insensitive: bool = True,
     ):
         # --- Validation ---
         if not isinstance(mapping, dict):
             raise TypeError("The 'mapping' argument must be a dictionary.")
+
+        self.unseen_value = unseen_value
         
+        # --- Process and validate patterns ---
+        # Process patterns here to be more efficient, avoiding reprocessing on every __call__.
+        self.processed_mapping: List[Tuple[str, Union[int, float]]] = []
         for pattern, value in mapping.items():
+            final_pattern = f"(?i){pattern}" if case_insensitive else pattern
+
+            # Validate the final pattern that will actually be used by Polars
             try:
-                re.compile(pattern)
+                re.compile(final_pattern)
             except re.error as e:
-                raise ValueError(f"Invalid regex pattern '{pattern}': {e}") from e
+                raise ValueError(f"Invalid regex pattern '{final_pattern}': {e}") from e
             if not isinstance(value, (int, float)):
                 raise TypeError(f"Mapping values must be int or float, but got {type(value)} for pattern '{pattern}'.")
-
-        self.mapping = mapping
-        self.unseen_value = unseen_value
+            
+            self.processed_mapping.append((final_pattern, value))
 
     def __call__(self, column: pl.Series) -> pl.Series:
         """
@@ -670,22 +860,20 @@ class RegexMapper:
             pl.Series: A new Series with strings mapped to numbers based on
                        the first matching regex pattern.
         """
-        # Ensure the column is treated as a string for matching
-        str_column = column.cast(pl.Utf8)
+        # pl.String is the modern alias for pl.Utf8
+        str_column = column.cast(pl.String)
 
-        # Build the when/then/otherwise chain from the inside out.
-        # Start with the final fallback value for non-matches.
+        # Start with the fallback value for non-matches.
         mapping_expr = pl.lit(self.unseen_value)
 
-        # Iterate through the mapping in reverse to construct the nested expression
-        for pattern, value in reversed(list(self.mapping.items())):
+        # Iterate through the processed mapping in reverse to construct the nested expression
+        for pattern, value in reversed(self.processed_mapping):
             mapping_expr = (
                 pl.when(str_column.str.contains(pattern))
                 .then(pl.lit(value))
                 .otherwise(mapping_expr)
             )
         
-        # Execute the complete expression chain and return the resulting Series
         return pl.select(mapping_expr).to_series()
 
 
