@@ -2,30 +2,30 @@ import numpy as np
 from pathlib import Path
 import xgboost as xgb
 import lightgbm as lgb
-from sklearn.ensemble import HistGradientBoostingRegressor
-from sklearn.base import ClassifierMixin
-from typing import Literal, Union, Tuple, Dict, Optional
+from typing import Literal, Union, Tuple, Dict, Optional, Any
 import pandas as pd
 from copy import deepcopy
-from .utilities import _script_info, threshold_binary_values, threshold_binary_values_batch, deserialize_object, list_files_by_extension, save_dataframe, make_fullpath, yield_dataframes_from_dir, sanitize_filename
+from .utilities import (
+    threshold_binary_values, 
+    threshold_binary_values_batch, 
+    deserialize_object,
+    yield_dataframes_from_dir)
+from .path_manager import sanitize_filename, make_fullpath, list_files_by_extension, list_csv_paths
 import torch
 from tqdm import trange
-import logging
 import matplotlib.pyplot as plt
 import seaborn as sns
-from collections import defaultdict
-
-# Configure logger
-logging.basicConfig(
-    level=logging.INFO,
-    format="[%(asctime)s] [%(levelname)s] - %(message)s",
-    datefmt="%Y-%m-%d %H:%M:%S"
-)
+from ._logger import _LOGGER
+from .keys import ModelSaveKeys
+from ._script_info import _script_info
+from .SQL import DatabaseManager
+from contextlib import nullcontext
 
 
 __all__ = [
     "ObjectiveFunction",
     "multiple_objective_functions_from_dir",
+    "parse_lower_upper_bounds",
     "run_pso",
     "plot_optimal_feature_distributions"
 ]
@@ -53,9 +53,9 @@ class ObjectiveFunction():
         self.is_hybrid = False if binary_features <= 0 else True
         self.use_noise = add_noise
         self._artifact = deserialize_object(trained_model_path, verbose=False, raise_on_error=True)
-        self.model = self._get_from_artifact('model')
-        self.feature_names: Optional[list[str]] = self._get_from_artifact('feature_names') # type: ignore
-        self.target_name: Optional[str] = self._get_from_artifact('target_name') # type: ignore
+        self.model = self._get_from_artifact(ModelSaveKeys.MODEL)
+        self.feature_names: Optional[list[str]] = self._get_from_artifact(ModelSaveKeys.FEATURES) # type: ignore
+        self.target_name: Optional[str] = self._get_from_artifact(ModelSaveKeys.TARGET) # type: ignore
         self.task = task
         self.check_model() # check for classification models and None values
         
@@ -122,7 +122,7 @@ class ObjectiveFunction():
             return features_array * noise
     
     def check_model(self):
-        if isinstance(self.model, ClassifierMixin) or isinstance(self.model, xgb.XGBClassifier) or isinstance(self.model, lgb.LGBMClassifier):
+        if isinstance(self.model, xgb.XGBClassifier) or isinstance(self.model, lgb.LGBMClassifier):
             raise ValueError(f"[Model Check Failed] ❌\nThe loaded model ({type(self.model).__name__}) is a Classifier.\nOptimization is not suitable for standard classification tasks.")
         if self.model is None:
             raise ValueError("Loaded model is None")
@@ -131,7 +131,7 @@ class ObjectiveFunction():
         if self._artifact is None:
             raise TypeError("Load model error")
         val = self._artifact.get(key)
-        if key == "feature_names":
+        if key == ModelSaveKeys.FEATURES:
             result = val if isinstance(val, list) and val else None
         else:
             result = val if val else None
@@ -170,6 +170,18 @@ def multiple_objective_functions_from_dir(directory: Union[str,Path], add_noise:
     return objective_functions, objective_function_names
 
 
+def parse_lower_upper_bounds(source: dict[str,tuple[Any,Any]]):
+    """
+    Parse lower and upper boundaries, returning 2 lists:
+    
+    `lower_bounds`, `upper_bounds`
+    """
+    lower = [low[0] for low in source.values()]
+    upper = [up[1] for up in source.values()]
+    
+    return lower, upper
+
+
 def _set_boundaries(lower_boundaries: list[float], upper_boundaries: list[float]):
     assert len(lower_boundaries) == len(upper_boundaries), "Lower and upper boundaries must have the same length."
     assert len(lower_boundaries) >= 1, "At least one boundary pair is required."
@@ -184,45 +196,73 @@ def _set_feature_names(size: int, names: Union[list[str], None]):
     else:
         assert len(names) == size, "List with feature names do not match the number of features"
         return names
-    
-
-def _save_results(*dicts, save_dir: Union[str,Path], target_name: str):
-    combined_dict = dict()
-    for single_dict in dicts:
-        combined_dict.update(single_dict)
-    
-    df = pd.DataFrame(combined_dict)
-    
-    save_dataframe(df=df, save_dir=save_dir, filename=f"Optimization_{target_name}")
 
 
-def _run_single_pso(objective_function: ObjectiveFunction, pso_args: dict, feature_names: list[str], target_name: str, random_state: int):
-    """Helper for a single PSO run."""
+def _save_result(result_dict: dict,
+                 save_format: Literal['csv', 'sqlite', 'both'],
+                 csv_path: Path,
+                 db_manager: Optional[DatabaseManager] = None,
+                 db_table_name: Optional[str] = None):
+    """
+    Handles saving a single result to CSV, SQLite, or both.
+    """
+    # Save to CSV
+    if save_format in ['csv', 'both']:
+        _save_or_append_to_csv(result_dict, csv_path)
+
+    # Save to SQLite
+    if save_format in ['sqlite', 'both']:
+        if db_manager and db_table_name:
+            db_manager.insert_row(db_table_name, result_dict)
+        else:
+            _LOGGER.warning("SQLite saving requested but db_manager or table_name not provided.")
+
+
+def _save_or_append_to_csv(data_dict: dict, save_path: Path):
+    """
+    Saves or appends a dictionary of data as a single row to a CSV file.
+
+    If the file doesn't exist, it creates it and writes the header.
+    If the file exists, it appends the new data without the header.
+    """
+    df_row = pd.DataFrame([data_dict])
+    
+    file_exists = save_path.exists()
+    
+    df_row.to_csv(
+        save_path, 
+        mode='a',              # 'a' for append mode
+        index=False,           # Don't write the DataFrame index
+        header=not file_exists # Write header only if file does NOT exist
+    )
+
+
+def _run_single_pso(objective_function: ObjectiveFunction, pso_args: dict, feature_names: list[str], target_name: str, random_state: int, save_format: Literal['csv', 'sqlite', 'both'], csv_path: Path, db_manager: Optional[DatabaseManager], db_table_name: str):
+    """Helper for a single PSO run that also handles saving."""
     pso_args.update({"seed": random_state})
     
     best_features, best_target, *_ = _pso(**pso_args)
     
-    # Flip best_target if maximization was used
     if objective_function.task == "maximization":
         best_target = -best_target
     
-    # Threshold binary features
     binary_number = objective_function.binary_features
     best_features_threshold = threshold_binary_values(best_features, binary_number)
     
-    # Name features and target
     best_features_named = {name: value for name, value in zip(feature_names, best_features_threshold)}
     best_target_named = {target_name: best_target}
     
+    # Save the result using the new helper
+    combined_dict = {**best_features_named, **best_target_named}
+    _save_result(combined_dict, save_format, csv_path, db_manager, db_table_name)
+    
     return best_features_named, best_target_named
 
 
-def _run_post_hoc_pso(objective_function: ObjectiveFunction, pso_args: dict, feature_names: list[str], target_name: str, repetitions: int):
-    """Helper for post-hoc PSO analysis."""
-    all_best_targets = []
-    all_best_features = [[] for _ in range(len(feature_names))]
-    
-    for _ in range(repetitions):
+def _run_post_hoc_pso(objective_function: ObjectiveFunction, pso_args: dict, feature_names: list[str], target_name: str, repetitions: int, save_format: Literal['csv', 'sqlite', 'both'], csv_path: Path, db_manager: Optional[DatabaseManager], db_table_name: str):
+    """Helper for post-hoc analysis that saves results incrementally."""
+    progress = trange(repetitions, desc="Post-Hoc PSO", unit="run")
+    for _ in progress:
         best_features, best_target, *_ = _pso(**pso_args)
         
         if objective_function.task == "maximization":
@@ -231,28 +271,25 @@ def _run_post_hoc_pso(objective_function: ObjectiveFunction, pso_args: dict, fea
         binary_number = objective_function.binary_features
         best_features_threshold = threshold_binary_values(best_features, binary_number)
         
-        for i, best_feature in enumerate(best_features_threshold):
-            all_best_features[i].append(best_feature)
-        all_best_targets.append(best_target)
-    
-    # Name features and target
-    all_best_features_named = {name: lst for name, lst in zip(feature_names, all_best_features)}
-    all_best_targets_named = {target_name: all_best_targets}
-    
-    return all_best_features_named, all_best_targets_named
+        result_dict = {name: value for name, value in zip(feature_names, best_features_threshold)}
+        result_dict[target_name] = best_target
+        
+        # Save each result incrementally
+        _save_result(result_dict, save_format, csv_path, db_manager, db_table_name)
 
 
 def run_pso(lower_boundaries: list[float], 
             upper_boundaries: list[float], 
             objective_function: ObjectiveFunction,
             save_results_dir: Union[str,Path],
+            save_format: Literal['csv', 'sqlite', 'both'] = 'csv',
             auto_binary_boundaries: bool=True,
             target_name: Union[str, None]=None, 
             feature_names: Union[list[str], None]=None,
             swarm_size: int=200, 
             max_iterations: int=3000,
             random_state: int=101,
-            post_hoc_analysis: Optional[int]=10) -> Tuple[Dict[str, float | list[float]], Dict[str, float | list[float]]]:
+            post_hoc_analysis: Optional[int]=10) -> Optional[Tuple[Dict[str, float], Dict[str, float]]]:
     """
     Executes Particle Swarm Optimization (PSO) to optimize a given objective function and saves the results as a CSV file.
 
@@ -266,6 +303,11 @@ def run_pso(lower_boundaries: list[float],
         A callable object encapsulating a tree-based regression model.
     save_results_dir : str | Path
         Directory path to save the results CSV file.
+    save_format : {'csv', 'sqlite', 'both'}, default 'csv'
+        The format for saving optimization results.
+        - 'csv': Saves results to a CSV file.
+        - 'sqlite': Saves results to an SQLite database file. ⚠️ If a database exists, new tables will be created using the target name.
+        - 'both': Saves results to both formats.
     auto_binary_boundaries : bool
         Use `ObjectiveFunction.binary_features` to append as many binary boundaries as needed to `lower_boundaries` and `upper_boundaries` automatically.
     target_name : str or None, optional
@@ -281,14 +323,11 @@ def run_pso(lower_boundaries: list[float],
 
     Returns
     -------
-    Tuple[Dict[str, float | list[float]], Dict[str, float | list[float]]]
-        If `post_hoc_analysis` is None, returns two dictionaries:
-            - feature_names: Feature values (after inverse scaling) that yield the best result.
-            - target_name: Best result obtained for the target variable.
-
-        If `post_hoc_analysis` is an integer, returns two dictionaries:
-            - feature_names: Lists of best feature values (after inverse scaling) for each repetition.
-            - target_name: List of best target values across repetitions.
+    Tuple[Dict[str, float], Dict[str, float]] or None
+        - If `post_hoc_analysis` is None, returns two dictionaries containing the
+          single best features and the corresponding target value.
+        - If `post_hoc_analysis` is active, results are streamed directly to a CSV file
+          and this function returns `None`.
 
     Notes
     -----
@@ -304,7 +343,7 @@ def run_pso(lower_boundaries: list[float],
     else:
         device = torch.device("cpu")
     
-    logging.info(f"Using device: '{device}'")
+    _LOGGER.info(f"👾 Using device: '{device}'")
     
     # set local deep copies to prevent in place list modification
     local_lower_boundaries = deepcopy(lower_boundaries)
@@ -313,8 +352,9 @@ def run_pso(lower_boundaries: list[float],
     # Append binary boundaries
     binary_number = objective_function.binary_features
     if auto_binary_boundaries and binary_number > 0:
-        local_lower_boundaries.extend([0] * binary_number)
-        local_upper_boundaries.extend([1] * binary_number)
+        # simplify binary search by constraining range
+        local_lower_boundaries.extend([0.45] * binary_number)
+        local_upper_boundaries.extend([0.55] * binary_number)
         
     # Set the total length of features
     size_of_features = len(local_lower_boundaries)
@@ -330,7 +370,25 @@ def run_pso(lower_boundaries: list[float],
     if target_name is None and objective_function.target_name is not None:
         target_name = objective_function.target_name
     if target_name is None:
-        target_name = "Target"
+        raise ValueError(f"'target' name was not provided and was not found in the .joblib object.")
+    
+    # --- Setup: Saving Infrastructure ---
+    sanitized_target_name = sanitize_filename(target_name)
+    save_dir_path = make_fullpath(save_results_dir, make=True, enforce="directory")
+    base_filename = f"Optimization_{sanitized_target_name}"
+    csv_path = save_dir_path / f"{base_filename}.csv"
+    db_path = save_dir_path / "Optimization.db"
+    db_table_name = f"{sanitized_target_name}"
+    
+    if save_format in ['sqlite', 'both']:
+        # Dynamically create the schema for the database table
+        schema = {name: "REAL" for name in names}
+        schema[target_name] = "REAL"
+        schema = {"result_id": "INTEGER PRIMARY KEY AUTOINCREMENT", **schema}
+        
+        # Create table
+        with DatabaseManager(db_path) as db:
+            db.create_table(db_table_name, schema)
         
     pso_arguments = {
             "func":objective_function,
@@ -342,17 +400,29 @@ def run_pso(lower_boundaries: list[float],
             "particle_output": False,
     }
     
-    # Dispatcher
-    if post_hoc_analysis is None or post_hoc_analysis <= 1:
-        features, target = _run_single_pso(objective_function, pso_arguments, names, target_name, random_state)
-    else:
-        features, target = _run_post_hoc_pso(objective_function, pso_arguments, names, target_name, post_hoc_analysis)
-    
-    # --- Save Results ---
-    save_results_path = make_fullpath(save_results_dir, make=True)
-    _save_results(features, target, save_dir=save_results_path, target_name=target_name)
-    
-    return features, target
+    # --- Dispatcher ---
+    # Use a real or dummy context manager to handle the DB connection cleanly
+    db_context = DatabaseManager(db_path) if save_format in ['sqlite', 'both'] else nullcontext()
+
+    with db_context as db_manager:
+        if post_hoc_analysis is None or post_hoc_analysis <= 1:
+            # --- Single Run Logic ---
+            features_dict, target_dict = _run_single_pso(
+                objective_function, pso_arguments, names, target_name, random_state,
+                save_format, csv_path, db_manager, db_table_name
+            )
+            _LOGGER.info(f"✅ Single optimization complete.")
+            return features_dict, target_dict
+        
+        else:
+            # --- Post-Hoc Analysis Logic ---
+            _LOGGER.info(f"🏁 Starting post-hoc analysis with {post_hoc_analysis} repetitions...")
+            _run_post_hoc_pso(
+                objective_function, pso_arguments, names, target_name, post_hoc_analysis,
+                save_format, csv_path, db_manager, db_table_name
+            )
+            _LOGGER.info("✅ Post-hoc analysis complete. Results saved.")
+            return None
 
 
 def _pso(func: ObjectiveFunction,
@@ -508,13 +578,13 @@ def _pso(func: ObjectiveFunction,
         return best_position, best_score
 
 
-def plot_optimal_feature_distributions(results_dir: Union[str, Path], save_dir: Union[str, Path], color_by_target: bool = True):
+def plot_optimal_feature_distributions(results_dir: Union[str, Path], save_dir: Union[str, Path]):
     """
     Analyzes optimization results and plots the distribution of optimal values for each feature.
 
-    This function can operate in two modes based on the `color_by_target` parameter:
-    1.  Aggregates all values for a feature into a single group and plots one overall distribution (histogram + KDE).
-    2.  Color-coded: Plots a separate, color-coded Kernel Density Estimate (KDE) for each source target, allowing for direct comparison on a single chart.
+    For features with more than two unique values, this function generates a color-coded 
+    Kernel Density Estimate (KDE) plot. For binary or constant features, it generates a bar plot
+    showing relative frequency.
 
     Parameters
     ----------
@@ -522,74 +592,69 @@ def plot_optimal_feature_distributions(results_dir: Union[str, Path], save_dir:
         The path to the directory containing the optimization result CSV files.
     save_dir : str or Path
         The directory where the output plots will be saved.
-    color_by_target : bool, optional
-        If True, generates comparative plots with distributions colored by their source target.
     """
-    mode = "Comparative (color-coded)" if color_by_target else "Aggregate"
-    logging.info(f"Starting analysis in '{mode}' mode from results in: '{results_dir}'")
-    
+    # Check results_dir and create output path
+    results_path = make_fullpath(results_dir)
     output_path = make_fullpath(save_dir, make=True)
-    all_files = list(yield_dataframes_from_dir(results_dir))
+    
+    # Check that the directory contains csv files
+    list_csv_paths(results_path, verbose=False)
 
-    if not all_files:
-        logging.warning("No data found. No plots will be generated.")
-        return
+    # --- Data Loading and Preparation ---
+    _LOGGER.info(f"📁 Starting analysis from results in: '{results_dir}'")
+    data_to_plot = []
+    for df, df_name in yield_dataframes_from_dir(results_path):
+        melted_df = df.iloc[:, :-1].melt(var_name='feature', value_name='value')
+        melted_df['target'] = df_name.replace("Optimization_", "")
+        data_to_plot.append(melted_df)
+    
+    long_df = pd.concat(data_to_plot, ignore_index=True)
+    features = long_df['feature'].unique()
+    _LOGGER.info(f"📂 Found data for {len(features)} features across {len(long_df['target'].unique())} targets. Generating plots...")
 
-    # --- MODE 1: Color-coded plots by target ---
-    if color_by_target:
-        data_to_plot = []
-        for df, df_name in all_files:
-            # Assumes last col is target, rest are features
-            melted_df = df.iloc[:, :-1].melt(var_name='feature', value_name='value')
-            # Sanitize target name for cleaner legend labels
-            melted_df['target'] = df_name.replace("Optimization_", "")
-            data_to_plot.append(melted_df)
+    # --- Plotting Loop ---
+    for feature_name in features:
+        plt.figure(figsize=(12, 7))
+        feature_df = long_df[long_df['feature'] == feature_name]
+
+        # Check if the feature is binary or constant
+        if feature_df['value'].nunique() <= 2:
+            # PLOT 1: For discrete values, calculate percentages and use a true bar plot.
+            # This ensures the X-axis is clean (e.g., just 0 and 1).
+            norm_df = (feature_df.groupby('target')['value']
+                       .value_counts(normalize=True)
+                       .mul(100)
+                       .rename('percent')
+                       .reset_index())
+            
+            ax = sns.barplot(data=norm_df, x='value', y='percent', hue='target')
+            
+            plt.title(f"Optimal Value Distribution for '{feature_name}'", fontsize=16)
+            plt.ylabel("Frequency (%)", fontsize=12)
+            ax.set_ylim(0, 100) # Set Y-axis from 0 to 100
+
+        else:
+            # PLOT 2: KDE plot for continuous values.
+            ax = sns.kdeplot(data=feature_df, x='value', hue='target',
+                             fill=True, alpha=0.1, warn_singular=False)
+
+            plt.title(f"Optimal Value Distribution for '{feature_name}'", fontsize=16)
+            plt.ylabel("Density", fontsize=12) # Y-axis is "Density" for KDE plots
+
+        # --- Common settings for both plot types ---
+        plt.xlabel("Feature Value", fontsize=12)
+        plt.grid(axis='y', alpha=0.5, linestyle='--')
         
-        long_df = pd.concat(data_to_plot, ignore_index=True)
-        features = long_df['feature'].unique()
-        logging.info(f"Found data for {len(features)} features across {len(long_df['target'].unique())} targets. Generating plots...")
+        legend = ax.get_legend()
+        if legend:
+            legend.set_title('Target')
 
-        for feature_name in features:
-            plt.figure(figsize=(12, 7))
-            feature_df = long_df[long_df['feature'] == feature_name]
-            
-            sns.kdeplot(data=feature_df, x='value', hue='target', fill=True, alpha=0.1)
-            
-            plt.title(f"Comparative Distribution for '{feature_name}'", fontsize=16)
-            plt.xlabel("Feature Value", fontsize=12)
-            plt.ylabel("Density", fontsize=12)
-            plt.grid(axis='y', alpha=0.5, linestyle='--')
-            plt.legend(title='Target')
+        sanitized_feature_name = sanitize_filename(feature_name)
+        plot_filename = output_path / f"Distribution_{sanitized_feature_name}.svg"
+        plt.savefig(plot_filename, bbox_inches='tight')
+        plt.close()
 
-            sanitized_feature_name = sanitize_filename(feature_name)
-            plot_filename = output_path / f"Comparative_{sanitized_feature_name}.svg"
-            plt.savefig(plot_filename, bbox_inches='tight')
-            plt.close()
-
-    # --- MODE 2: Aggregate plot ---
-    else:
-        feature_distributions = defaultdict(list)
-        for df, _ in all_files:
-            feature_columns = df.iloc[:, :-1]
-            for feature_name in feature_columns:
-                feature_distributions[feature_name].extend(df[feature_name].tolist())
-        
-        logging.info(f"Found data for {len(feature_distributions)} features. Generating plots...")
-        for feature_name, values in feature_distributions.items():
-            plt.figure(figsize=(12, 7))
-            sns.histplot(x=values, kde=True, bins='auto', stat="density")
-            
-            plt.title(f"Aggregate Distribution for '{feature_name}'", fontsize=16)
-            plt.xlabel("Feature Value", fontsize=12)
-            plt.ylabel("Density", fontsize=12)
-            plt.grid(axis='y', alpha=0.5, linestyle='--')
-
-            sanitized_feature_name = sanitize_filename(feature_name)
-            plot_filename = output_path / f"Aggregate_{sanitized_feature_name}.svg"
-            plt.savefig(plot_filename, bbox_inches='tight')
-            plt.close()
-
-    logging.info(f"✅ All plots saved successfully to: {output_path}")
+    _LOGGER.info(f"✅ All plots saved successfully to: '{output_path}'")
 
 
 def info():
