@@ -3,25 +3,23 @@ from pandas.api.types import is_numeric_dtype
 import numpy as np
 import matplotlib.pyplot as plt
 import seaborn as sns
-from IPython import get_ipython
-from IPython.display import clear_output
-import time
 from typing import Union, Literal, Dict, Tuple, List, Optional
 from pathlib import Path
-from .utilities import sanitize_filename, _script_info, make_fullpath
+from .path_manager import sanitize_filename, make_fullpath
+from ._script_info import _script_info
 import re
 
 
 # Keep track of all available tools, show using `info()`
 __all__ = [
     "summarize_dataframe",
+    "drop_constant_columns",
     "drop_rows_with_missing_data",
-    "split_features_targets", 
     "show_null_columns",
     "drop_columns_with_missing_data",
+    "split_features_targets", 
     "split_continuous_binary", 
     "plot_correlation_heatmap", 
-    "check_value_distributions", 
     "plot_value_distributions", 
     "clip_outliers_single", 
     "clip_outliers_multi",
@@ -61,6 +59,53 @@ def summarize_dataframe(df: pd.DataFrame, round_digits: int = 2):
     return summary
 
 
+def drop_constant_columns(df: pd.DataFrame, verbose: bool = True) -> pd.DataFrame:
+    """
+    Removes columns from a pandas DataFrame that contain only a single unique 
+    value or are entirely null/NaN.
+
+    This utility is useful for cleaning data by removing constant features that 
+    have no predictive value.
+
+    Args:
+        df (pd.DataFrame): 
+            The pandas DataFrame to clean.
+        verbose (bool): 
+            If True, prints the names of the columns that were dropped. 
+            Defaults to True.
+
+    Returns:
+        pd.DataFrame: 
+            A new DataFrame with the constant columns removed.
+    """
+    if not isinstance(df, pd.DataFrame):
+        raise TypeError("Input must be a pandas DataFrame.")
+
+    original_columns = set(df.columns)
+    cols_to_keep = []
+
+    for col_name in df.columns:
+        column = df[col_name]
+        
+        # We can apply this logic to all columns or only focus on numeric ones.
+        # if not is_numeric_dtype(column):
+        #     cols_to_keep.append(col_name)
+        #     continue
+        
+        # Keep a column if it has more than one unique value (nunique ignores NaNs by default)
+        if column.nunique(dropna=True) > 1:
+            cols_to_keep.append(col_name)
+
+    dropped_columns = original_columns - set(cols_to_keep)
+    if verbose:
+        print(f"🧹 Dropped {len(dropped_columns)} constant columns.")
+        if dropped_columns:
+            for dropped_column in dropped_columns:
+                print(f"    {dropped_column}")
+
+    return df[cols_to_keep]
+
+
 def drop_rows_with_missing_data(df: pd.DataFrame, targets: Optional[list[str]], threshold: float = 0.7) -> pd.DataFrame:
     """
     Drops rows from the DataFrame using a two-stage strategy:
@@ -80,22 +125,24 @@ def drop_rows_with_missing_data(df: pd.DataFrame, targets: Optional[list[str]],
 
     # Stage 1: Drop rows with all target columns missing
     if targets is not None:
-        target_na = df_clean[targets].isnull().all(axis=1)
+        # validate targets
+        valid_targets = _validate_columns(df_clean, targets)
+        target_na = df_clean[valid_targets].isnull().all(axis=1)
         if target_na.any():
             print(f"🧹 Dropping {target_na.sum()} rows with all target columns missing.")
             df_clean = df_clean[~target_na]
         else:
             print("✅ No rows with all targets missing.")
     else:
-        targets = []
+        valid_targets = []
 
     # Stage 2: Drop rows based on feature column missing values
-    feature_cols = [col for col in df_clean.columns if col not in targets]
+    feature_cols = [col for col in df_clean.columns if col not in valid_targets]
     if feature_cols:
         feature_na_frac = df_clean[feature_cols].isnull().mean(axis=1)
         rows_to_drop = feature_na_frac[feature_na_frac > threshold].index
         if len(rows_to_drop) > 0:
-            print(f"📉 Dropping {len(rows_to_drop)} rows with more than {threshold*100:.0f}% missing feature data.")
+            print(f"🧹 Dropping {len(rows_to_drop)} rows with more than {threshold*100:.0f}% missing feature data.")
             df_clean = df_clean.drop(index=rows_to_drop)
         else:
             print(f"✅ No rows exceed the {threshold*100:.0f}% missing feature data threshold.")
@@ -105,30 +152,6 @@ def drop_rows_with_missing_data(df: pd.DataFrame, targets: Optional[list[str]],
     return df_clean
 
 
-def split_features_targets(df: pd.DataFrame, targets: list[str]):
-    """
-    Splits a DataFrame's columns into features and targets.
-
-    Args:
-        df (pd.DataFrame): Pandas DataFrame containing the dataset.
-        targets (list[str]): List of column names to be treated as target variables.
-
-    Returns:
-        tuple: A tuple containing:
-            - pd.DataFrame: Features dataframe.
-            - pd.DataFrame: Targets dataframe.
-
-    Prints:
-        - Shape of the original dataframe.
-        - Shape of the features dataframe.
-        - Shape of the targets dataframe.
-    """
-    df_targets = df[targets]
-    df_features = df.drop(columns=targets)
-    print(f"Original shape: {df.shape}\nFeatures shape: {df_features.shape}\nTargets shape: {df_targets.shape}")
-    return df_features, df_targets
-
-
 def show_null_columns(df: pd.DataFrame, round_digits: int = 2):
     """
     Displays a table of columns with missing values, showing both the count and
@@ -157,7 +180,7 @@ def show_null_columns(df: pd.DataFrame, round_digits: int = 2):
     return null_summary
 
 
-def drop_columns_with_missing_data(df: pd.DataFrame, threshold: float = 0.7, show_nulls_after: bool = True) -> pd.DataFrame:
+def drop_columns_with_missing_data(df: pd.DataFrame, threshold: float = 0.7, show_nulls_after: bool = True, skip_columns: Optional[List[str]]=None) -> pd.DataFrame:
     """
     Drops columns with more than `threshold` fraction of missing values.
 
@@ -165,11 +188,22 @@ def drop_columns_with_missing_data(df: pd.DataFrame, threshold: float = 0.7, sho
         df (pd.DataFrame): The input DataFrame.
         threshold (float): Fraction of missing values above which columns are dropped.
         show_nulls_after (bool): Prints `show_null_columns` after dropping columns. 
+        skip_columns (list[str] | None): If given, these columns wont be included in the drop process. 
 
     Returns:
         pd.DataFrame: A new DataFrame without the dropped columns.
     """
-    missing_fraction = df.isnull().mean()
+    # If skip_columns is provided, create a list of columns to check.
+    # Otherwise, check all columns.
+    cols_to_check = df.columns
+    if skip_columns:
+        # Use set difference for efficient exclusion
+        cols_to_check = df.columns.difference(skip_columns)
+
+    # Calculate the missing fraction only on the columns to be checked
+    missing_fraction = df[cols_to_check].isnull().mean()
+    
+    
     cols_to_drop = missing_fraction[missing_fraction > threshold].index
 
     if len(cols_to_drop) > 0:
@@ -186,6 +220,31 @@ def drop_columns_with_missing_data(df: pd.DataFrame, threshold: float = 0.7, sho
         return df
 
 
+def split_features_targets(df: pd.DataFrame, targets: list[str]):
+    """
+    Splits a DataFrame's columns into features and targets.
+
+    Args:
+        df (pd.DataFrame): Pandas DataFrame containing the dataset.
+        targets (list[str]): List of column names to be treated as target variables.
+
+    Returns:
+        tuple: A tuple containing:
+            - pd.DataFrame: Features dataframe.
+            - pd.DataFrame: Targets dataframe.
+
+    Prints:
+        - Shape of the original dataframe.
+        - Shape of the features dataframe.
+        - Shape of the targets dataframe.
+    """
+    valid_targets = _validate_columns(df, targets)
+    df_targets = df[valid_targets]
+    df_features = df.drop(columns=valid_targets)
+    print(f"Original shape: {df.shape}\nFeatures shape: {df_features.shape}\nTargets shape: {df_targets.shape}")
+    return df_features, df_targets
+
+
 def split_continuous_binary(df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
     """
     Split DataFrame into two DataFrames: one with continuous columns, one with binary columns.
@@ -289,69 +348,12 @@ def plot_correlation_heatmap(df: pd.DataFrame,
         full_path = save_path / plot_title
         
         plt.savefig(full_path, bbox_inches="tight", format='svg')
-        print(f"Saved correlation heatmap: '{plot_title}.svg'")
+        print(f"Saved correlation heatmap: '{plot_title}'")
     
     plt.show()
     plt.close()
 
 
-def check_value_distributions(df: pd.DataFrame, view_frequencies: bool=True, bin_threshold: int=10, skip_cols_with_key: Union[str, None]=None):
-    """
-    Analyzes value counts for each column in a DataFrame, optionally plots distributions, 
-    and saves them as .png files in the specified directory.
-
-    Args:
-        df (pd.DataFrame): The dataset to analyze.
-        view_frequencies (bool): Print relative frequencies instead of value counts.
-        bin_threshold (int): Threshold of unique values to start using bins.
-        skip_cols_with_key (str | None): Skip column names containing the key. If None, don't skip any column.
-    
-    Notes:
-        - Binning is adaptive: if quantile binning results in ≤ 2 unique bins, raw values are used instead.
-    """
-    # cherry-pick columns
-    if skip_cols_with_key is not None:
-        columns = [col for col in df.columns if skip_cols_with_key not in col]
-    else:
-        columns = df.columns.to_list()
-    
-    for col in columns:
-        if _is_notebook():
-            clear_output(wait=False)
-        if pd.api.types.is_numeric_dtype(df[col]) and df[col].nunique() > bin_threshold:
-            bins_number = 10
-            binned = pd.qcut(df[col], q=bins_number, duplicates='drop')
-            while binned.nunique() <= 2:
-                bins_number -= 1
-                binned = pd.qcut(df[col], q=bins_number, duplicates='drop')
-                if bins_number <= 2:
-                    break
-            
-            if binned.nunique() <= 2:
-                view_std = df[col].value_counts(ascending=False)
-            else:
-                view_std = binned.value_counts(sort=False)
-            
-        else:
-            view_std = df[col].value_counts(ascending=False)
-
-        view_std.name = col
-
-        # unlikely scenario where the series is empty
-        if view_std.sum() == 0:
-            view_freq = view_std
-        else:
-            view_freq = view_std / view_std.sum()
-        # view_freq = df[col].value_counts(normalize=True, bins=10)  # relative percentages
-        view_freq.name = col
-
-        # Print value counts
-        print(view_freq if view_frequencies else view_std)
-        
-        time.sleep(1)
-        user_input_ = input("Press enter to continue")
-
-
 def plot_value_distributions(df: pd.DataFrame, save_dir: Union[str, Path], bin_threshold: int=10, skip_cols_with_key: Union[str, None]=None):
     """
     Plots and saves the value distributions for all (or selected) columns in a DataFrame, 
@@ -587,14 +589,14 @@ def standardize_percentages(
     Standardizes numeric columns containing mixed-format percentages.
 
     This function cleans columns where percentages might be entered as whole
-    numbers (e.g., 55) or as proportions (e.g., 0.55). It assumes values
+    numbers (55) and as proportions (0.55). It assumes values
     between 0 and 1 are proportions and multiplies them by 100.
 
     Args:
         df (pd.Dataframe): The input pandas DataFrame.
         columns (list[str]): A list of column names to standardize.
         treat_one_as_proportion (bool):
-            - If True (default): The value `1` is treated as a proportion and converted to `100`.
+            - If True (default): The value `1` is treated as a proportion and converted to `100%`.
             - If False: The value `1` is treated as `1%`.
         round_digits (int): The number of decimal places to round the final result to.
 
@@ -643,8 +645,9 @@ def standardize_percentages(
     return df_copy
 
 
-def _is_notebook():
-    return get_ipython() is not None
+def _validate_columns(df: pd.DataFrame, columns: list[str]):
+    valid_columns = [column for column in columns if column in df.columns]
+    return valid_columns
 
 
 def info():
