@@ -1,6 +1,6 @@
 import pandas as pd
 import numpy as np
-import seaborn as sns
+import seaborn # Use plot styling
 import matplotlib.pyplot as plt
 from matplotlib.colors import Colormap
 from matplotlib import rcdefaults
@@ -11,7 +11,6 @@ from typing import Literal, Union, Optional, Iterator, Tuple
 from imblearn.over_sampling import ADASYN, SMOTE, RandomOverSampler
 from imblearn.under_sampling import RandomUnderSampler
 
-from sklearn.ensemble import HistGradientBoostingClassifier, HistGradientBoostingRegressor
 import xgboost as xgb
 import lightgbm as lgb
 
@@ -19,7 +18,11 @@ from sklearn.model_selection import train_test_split
 from sklearn.metrics import accuracy_score, classification_report, ConfusionMatrixDisplay, mean_absolute_error, mean_squared_error, r2_score, roc_curve, roc_auc_score
 import shap
 
-from .utilities import yield_dataframes_from_dir, sanitize_filename, _script_info, serialize_object, make_fullpath
+from .utilities import yield_dataframes_from_dir, serialize_object
+from .path_manager import sanitize_filename, make_fullpath
+from ._script_info import _script_info
+from .keys import ModelSaveKeys
+from ._logger import _LOGGER
 
 import warnings # Ignore warnings 
 warnings.filterwarnings('ignore', category=DeprecationWarning)
@@ -37,7 +40,7 @@ __all__ = [
     "evaluate_model_regression",
     "get_shap_values",
     "train_test_pipeline",
-    "run_ensemble_pipeline"
+    "run_ensemble_pipeline",
 ]
 
 ## Type aliases
@@ -76,7 +79,7 @@ def dataset_yielder(
 class RegressionTreeModels:
     """
     A factory class for creating and configuring multiple gradient boosting regression models
-    with unified hyperparameters. This includes XGBoost, LightGBM, and HistGradientBoostingRegressor.
+    with unified hyperparameters. This includes XGBoost and LightGBM.
     
     Use the `__call__`, `()` method.
 
@@ -106,12 +109,6 @@ class RegressionTreeModels:
     colsample_bytree : float [0.3 - 1.0]
         Fraction of features per tree; useful for regularization (used by XGBoost and LightGBM).
 
-    min_samples_leaf : int [10 - 100]
-        Minimum samples per leaf; higher = less overfitting (used in HistGB).
-
-    max_iter : int [100 - 2000]
-        Maximum number of iterations (used in HistGB).
-
     min_child_weight : float [0.1 - 10.0]
         Minimum sum of instance weight (hessian) needed in a child; larger values make the algorithm more conservative (used in XGBoost).
 
@@ -125,20 +122,19 @@ class RegressionTreeModels:
         Minimum number of data points in a leaf; increasing may prevent overfitting (used in LightGBM).
     """
     def __init__(self, 
-                 random_state: int = 101,
-                 learning_rate: float = 0.005,
-                 L1_regularization: float = 1.0,
-                 L2_regularization: float = 1.0,
-                 n_estimators: int = 1000,
-                 max_depth: int = 8,
-                 subsample: float = 0.8,
-                 colsample_bytree: float = 0.8,
-                 min_samples_leaf: int = 50,
-                 max_iter: int = 1000,
-                 min_child_weight: float = 3.0,
-                 gamma: float = 1.0,
-                 num_leaves: int = 31,
-                 min_data_in_leaf: int = 40):
+             random_state: int = 101,
+             learning_rate: float = 0.005,
+             L1_regularization: float = 1.0,
+             L2_regularization: float = 1.0,
+             n_estimators: int = 1000,
+             max_depth: int = 8,
+             subsample: float = 0.8,
+             colsample_bytree: float = 0.8,
+             min_child_weight: float = 3.0,
+             gamma: float = 1.0,
+             num_leaves: int = 31,
+             min_data_in_leaf: int = 40):
+        
         # General config
         self.random_state = random_state
         self.lr = learning_rate
@@ -160,16 +156,11 @@ class RegressionTreeModels:
         self.num_leaves = num_leaves
         self.min_data_in_leaf = min_data_in_leaf
 
-        # HistGB specific
-        self.max_iter = max_iter
-        self.min_samples_leaf = min_samples_leaf
-
     def __call__(self) -> dict[str, object]:
         """
         Returns a dictionary with new instances of:
             - "XGBoost": XGBRegressor
             - "LightGBM": LGBMRegressor
-            - "HistGB": HistGradientBoostingRegressor
         """
         # XGBoost Regressor
         xgb_model = xgb.XGBRegressor(
@@ -204,23 +195,9 @@ class RegressionTreeModels:
             min_data_in_leaf=self.min_data_in_leaf
         )
 
-        # HistGradientBoosting Regressor
-        hist_model = HistGradientBoostingRegressor(
-            max_iter=self.max_iter,
-            learning_rate=self.lr,
-            max_depth=self.max_depth,
-            min_samples_leaf=self.min_samples_leaf,
-            random_state=self.random_state,
-            l2_regularization=self.L2,
-            scoring='neg_mean_squared_error',
-            early_stopping=True,
-            validation_fraction=0.1
-        )
-
         return {
             "XGBoost": xgb_model,
-            "LightGBM": lgb_model,
-            "HistGB": hist_model
+            "LightGBM": lgb_model
         }
     
     def __str__(self):
@@ -230,7 +207,7 @@ class RegressionTreeModels:
 class ClassificationTreeModels:
     """
     A factory class for creating and configuring multiple gradient boosting classification models
-    with unified hyperparameters. This includes: XGBoost, LightGBM, and HistGradientBoostingClassifier.
+    with unified hyperparameters. This includes XGBoost and LightGBM.
     
     Use the `__call__`, `()` method.
 
@@ -260,12 +237,6 @@ class ClassificationTreeModels:
     colsample_bytree : float [0.3 - 1.0]
         Fraction of features per tree; useful for regularization (used by XGBoost and LightGBM).
 
-    min_samples_leaf : int [10 - 100]
-        Minimum number of samples required to be at a leaf node; higher = less overfitting (used in HistGB).
-
-    max_iter : int [100 - 2000]
-        Maximum number of boosting iteration (used in HistGB).
-
     min_child_weight : float [0.1 - 10.0]
         Minimum sum of instance weight (Hessian) in a child node; larger values make the algorithm more conservative (used in XGBoost).
 
@@ -284,20 +255,19 @@ class ClassificationTreeModels:
         Indicates whether to apply class balancing strategies internally. Can be overridden at runtime via the `__call__` method.
     """
     def __init__(self,
-                 random_state: int = 101,
-                 learning_rate: float = 0.005,
-                 L1_regularization: float = 1.0,
-                 L2_regularization: float = 1.0,
-                 n_estimators: int = 1000,
-                 max_depth: int = 8,
-                 subsample: float = 0.8,
-                 colsample_bytree: float = 0.8,
-                 min_samples_leaf: int = 50,
-                 max_iter: int = 1000,
-                 min_child_weight: float = 3.0,
-                 gamma: float = 1.0,
-                 num_leaves: int = 31,
-                 min_data_in_leaf: int = 40):
+             random_state: int = 101,
+             learning_rate: float = 0.005,
+             L1_regularization: float = 1.0,
+             L2_regularization: float = 1.0,
+             n_estimators: int = 1000,
+             max_depth: int = 8,
+             subsample: float = 0.8,
+             colsample_bytree: float = 0.8,
+             min_child_weight: float = 3.0,
+             gamma: float = 1.0,
+             num_leaves: int = 31,
+             min_data_in_leaf: int = 40):
+        
         # General config
         self.random_state = random_state
         self.lr = learning_rate
@@ -322,16 +292,11 @@ class ClassificationTreeModels:
         self.num_leaves = num_leaves
         self.min_data_in_leaf = min_data_in_leaf
 
-        # HistGB specific
-        self.max_iter = max_iter
-        self.min_samples_leaf = min_samples_leaf
-
     def __call__(self, use_model_balance: Optional[bool]=None) -> dict[str, object]:
         """
         Returns a dictionary with new instances of:
             - "XGBoost": XGBClassifier
             - "LightGBM": LGBMClassifier
-            - "HistGB": HistGradientBoostingClassifier
         """
         if use_model_balance is not None:
             self.use_model_balance = use_model_balance
@@ -371,24 +336,9 @@ class ClassificationTreeModels:
             class_weight='balanced' if self.use_model_balance else None
         )
 
-        # HistGradientBoosting Classifier
-        hist_model = HistGradientBoostingClassifier(
-            max_iter=self.max_iter,
-            learning_rate=self.lr,
-            max_depth=self.max_depth,
-            min_samples_leaf=self.min_samples_leaf,
-            random_state=self.random_state,
-            l2_regularization=self.L2,
-            early_stopping=True,
-            validation_fraction=0.1,
-            class_weight='balanced' if self.use_model_balance else None,
-            scoring='balanced_accuracy' if self.use_model_balance else 'loss'
-        )
-
         return {
             "XGBoost": xgb_model,
-            "LightGBM": lgb_model,
-            "HistGB": hist_model
+            "LightGBM": lgb_model
         }
         
     def __str__(self):
@@ -438,16 +388,16 @@ def dataset_pipeline(df_features: pd.DataFrame, df_target: pd.Series, task: Task
     '''
     #DEBUG
     if debug:
-        print(f"Split Dataframes Shapes - Features DF: {df_features.shape}, Target DF: {df_target.shape}")
+        _LOGGER.info(f"Split Dataframes Shapes - Features DF: {df_features.shape}, Target DF: {df_target.shape}")
         unique_values = df_target.unique()  # Get unique values for the target column
-        print(f"\tUnique values for '{df_target.name}': {unique_values}")
+        _LOGGER.info(f"\tUnique values for '{df_target.name}': {unique_values}")
     
     #Train test split
     X_train, X_test, y_train, y_test = _split_data(features=df_features, target=df_target, test_size=test_size, random_state=random_state, task=task)
     
     #DEBUG
     if debug:
-        print(f"Shapes after train test split - X_train: {X_train.shape}, y_train: {y_train.shape}, X_test: {X_test.shape}, y_test: {y_test.shape}")
+        _LOGGER.info(f"Shapes after train test split - X_train: {X_train.shape}, y_train: {y_train.shape}, X_test: {X_test.shape}, y_test: {y_test.shape}")
     
  
     # Resample
@@ -458,7 +408,7 @@ def dataset_pipeline(df_features: pd.DataFrame, df_target: pd.Series, task: Task
     
     #DEBUG
     if debug:
-        print(f"Shapes after resampling - X_train: {X_train_oversampled.shape}, y_train: {y_train_oversampled.shape}, X_test: {X_test.shape}, y_test: {y_test.shape}")
+        _LOGGER.info(f"Shapes after resampling - X_train: {X_train_oversampled.shape}, y_train: {y_train_oversampled.shape}, X_test: {X_test.shape}, y_test: {y_test.shape}")
     
     return X_train_oversampled, y_train_oversampled, X_test, y_test
 
@@ -485,8 +435,10 @@ def _save_model(trained_model, model_name: str, target_name:str, feature_names:
     #Sanitize filenames to save
     sanitized_target_name = sanitize_filename(target_name)
     filename = f"{model_name}_{sanitized_target_name}"
-    to_save = {'model': trained_model, 'feature_names': feature_names, 'target_name':target_name}
-    
+    to_save = {ModelSaveKeys.MODEL: trained_model, 
+               ModelSaveKeys.FEATURES: feature_names,
+               ModelSaveKeys.TARGET: target_name}
+
     serialize_object(obj=to_save, save_dir=save_directory, filename=filename, verbose=False, raise_on_error=True)
 
 # function to evaluate the model and save metrics (Classification)
@@ -570,7 +522,7 @@ def evaluate_model_classification(
 
     fig.tight_layout()
     fig_path = save_path / f"Confusion_Matrix_{sanitized_target_name}.svg"
-    fig.savefig(fig_path, format="svg", bbox_inches="tight")
+    fig.savefig(fig_path, format="svg", bbox_inches="tight") # type: ignore
     plt.close(fig)
 
     return y_pred
@@ -614,8 +566,8 @@ def plot_roc_curve(
     # Determine predicted probabilities
     if isinstance(probabilities_or_model, np.ndarray):
         # Input is already probabilities
-        if probabilities_or_model.ndim == 2:
-            y_score = probabilities_or_model[:, 1]
+        if probabilities_or_model.ndim == 2: # type: ignore
+            y_score = probabilities_or_model[:, 1] # type: ignore
         else:
             y_score = probabilities_or_model
             
@@ -654,7 +606,7 @@ def plot_roc_curve(
     save_path = make_fullpath(save_directory, make=True)
     sanitized_target_name = sanitize_filename(target_name)
     full_save_path = save_path / f"ROC_{sanitized_target_name}.svg"
-    fig.savefig(full_save_path, bbox_inches="tight", format="svg")
+    fig.savefig(full_save_path, bbox_inches="tight", format="svg") # type: ignore
 
     return fig
 
@@ -864,7 +816,7 @@ def train_test_pipeline(model, model_name: str, dataset_id: str, task: TaskType,
     print(f"\tTraining model: {model_name} for Target: {target_name}...")
     trained_model = _train_model(model=model, train_features=train_features, train_target=train_target)
     if debug:
-        print(f"Trained model object: {type(trained_model)}")
+        _LOGGER.info(f"Trained model object: {type(trained_model)}")
     local_save_directory = _local_directories(model_name=model_name, dataset_id=dataset_id, save_dir=save_dir)
     
     if save_model:
@@ -885,11 +837,11 @@ def train_test_pipeline(model, model_name: str, dataset_id: str, task: TaskType,
     else:
         raise ValueError(f"Unrecognized task '{task}' for model training,")
     if debug:
-        print(f"Predicted vector: {type(y_pred)} with shape: {y_pred.shape}")
+        _LOGGER.info(f"Predicted vector: {type(y_pred)} with shape: {y_pred.shape}")
     
     get_shap_values(model=trained_model, model_name=model_name, save_dir=local_save_directory,
                     features_to_explain=train_features, feature_names=feature_names, target_name=target_name, task=task)
-    # print("\t...done.")
+    
     return trained_model, y_pred
 
 ###### 5. Execution ######
@@ -902,7 +854,7 @@ def run_ensemble_pipeline(datasets_dir: Union[str,Path], save_dir: Union[str,Pat
     elif isinstance(model_object, ClassificationTreeModels):
         task = "classification"
         if handle_classification_imbalance is None:
-            print("⚠️ No method to handle classification class imbalance has been selected. Datasets are assumed to be balanced.")
+            _LOGGER.warning("⚠️ No method to handle classification class imbalance has been selected. Datasets are assumed to be balanced.")
         elif handle_classification_imbalance == "by_model":
             model_object.use_model_balance = True
         else:
@@ -914,6 +866,7 @@ def run_ensemble_pipeline(datasets_dir: Union[str,Path], save_dir: Union[str,Pat
     datasets_path = make_fullpath(datasets_dir)
     save_path = make_fullpath(save_dir, make=True)
     
+    _LOGGER.info("🏁 Training starting...")
     #Yield imputed dataset
     for dataframe, dataframe_name in yield_dataframes_from_dir(datasets_path):
         #Yield features dataframe and target dataframe
@@ -931,7 +884,8 @@ def run_ensemble_pipeline(datasets_dir: Union[str,Path], save_dir: Union[str,Pat
                                     test_features=X_test, test_target=y_test,
                                     feature_names=feature_names,target_name=target_name,
                                     debug=debug, save_dir=save_path, save_model=save_model)
-    print("\n✅ Training and evaluation complete.")
+
+    _LOGGER.info("✅ Training and evaluation complete.")
 
 
 def info():
